# -*- coding: utf-8 -*-
"""Implements protocols on top of lower level libraries to talk to instruments.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
